# Copyright 2019 Richard Mitchell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Based on portions of https://github.com/codetheweb/tuyapi/
#
# MIT License
#
# Copyright (c) 2017 Max Isom
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
import base64
import contextlib
import json
import logging
import struct
import time
from collections.abc import Callable
from typing import Any, Self

from cryptography.hazmat.backends.openssl import backend as openssl_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.hashes import MD5, Hash
from cryptography.hazmat.primitives.padding import PKCS7

from .exceptions import (
    ConnectionException,
    ConnectionFailed,
    ConnectionTimeoutException,
    InvalidKey,
    InvalidMessage,
    MessageDecodeFailed,
)

_LOGGER = logging.getLogger(__name__)

MESSAGE_PREFIX_FORMAT = ">IIII"
MESSAGE_SUFFIX_FORMAT = ">II"
MAGIC_PREFIX = 0x000055AA
MAGIC_SUFFIX = 0x0000AA55
MAGIC_SUFFIX_BYTES = struct.pack(">I", MAGIC_SUFFIX)
CRC_32_TABLE = [
    0x00000000,
    0x77073096,
    0xEE0E612C,
    0x990951BA,
    0x076DC419,
    0x706AF48F,
    0xE963A535,
    0x9E6495A3,
    0x0EDB8832,
    0x79DCB8A4,
    0xE0D5E91E,
    0x97D2D988,
    0x09B64C2B,
    0x7EB17CBD,
    0xE7B82D07,
    0x90BF1D91,
    0x1DB71064,
    0x6AB020F2,
    0xF3B97148,
    0x84BE41DE,
    0x1ADAD47D,
    0x6DDDE4EB,
    0xF4D4B551,
    0x83D385C7,
    0x136C9856,
    0x646BA8C0,
    0xFD62F97A,
    0x8A65C9EC,
    0x14015C4F,
    0x63066CD9,
    0xFA0F3D63,
    0x8D080DF5,
    0x3B6E20C8,
    0x4C69105E,
    0xD56041E4,
    0xA2677172,
    0x3C03E4D1,
    0x4B04D447,
    0xD20D85FD,
    0xA50AB56B,
    0x35B5A8FA,
    0x42B2986C,
    0xDBBBC9D6,
    0xACBCF940,
    0x32D86CE3,
    0x45DF5C75,
    0xDCD60DCF,
    0xABD13D59,
    0x26D930AC,
    0x51DE003A,
    0xC8D75180,
    0xBFD06116,
    0x21B4F4B5,
    0x56B3C423,
    0xCFBA9599,
    0xB8BDA50F,
    0x2802B89E,
    0x5F058808,
    0xC60CD9B2,
    0xB10BE924,
    0x2F6F7C87,
    0x58684C11,
    0xC1611DAB,
    0xB6662D3D,
    0x76DC4190,
    0x01DB7106,
    0x98D220BC,
    0xEFD5102A,
    0x71B18589,
    0x06B6B51F,
    0x9FBFE4A5,
    0xE8B8D433,
    0x7807C9A2,
    0x0F00F934,
    0x9609A88E,
    0xE10E9818,
    0x7F6A0DBB,
    0x086D3D2D,
    0x91646C97,
    0xE6635C01,
    0x6B6B51F4,
    0x1C6C6162,
    0x856530D8,
    0xF262004E,
    0x6C0695ED,
    0x1B01A57B,
    0x8208F4C1,
    0xF50FC457,
    0x65B0D9C6,
    0x12B7E950,
    0x8BBEB8EA,
    0xFCB9887C,
    0x62DD1DDF,
    0x15DA2D49,
    0x8CD37CF3,
    0xFBD44C65,
    0x4DB26158,
    0x3AB551CE,
    0xA3BC0074,
    0xD4BB30E2,
    0x4ADFA541,
    0x3DD895D7,
    0xA4D1C46D,
    0xD3D6F4FB,
    0x4369E96A,
    0x346ED9FC,
    0xAD678846,
    0xDA60B8D0,
    0x44042D73,
    0x33031DE5,
    0xAA0A4C5F,
    0xDD0D7CC9,
    0x5005713C,
    0x270241AA,
    0xBE0B1010,
    0xC90C2086,
    0x5768B525,
    0x206F85B3,
    0xB966D409,
    0xCE61E49F,
    0x5EDEF90E,
    0x29D9C998,
    0xB0D09822,
    0xC7D7A8B4,
    0x59B33D17,
    0x2EB40D81,
    0xB7BD5C3B,
    0xC0BA6CAD,
    0xEDB88320,
    0x9ABFB3B6,
    0x03B6E20C,
    0x74B1D29A,
    0xEAD54739,
    0x9DD277AF,
    0x04DB2615,
    0x73DC1683,
    0xE3630B12,
    0x94643B84,
    0x0D6D6A3E,
    0x7A6A5AA8,
    0xE40ECF0B,
    0x9309FF9D,
    0x0A00AE27,
    0x7D079EB1,
    0xF00F9344,
    0x8708A3D2,
    0x1E01F268,
    0x6906C2FE,
    0xF762575D,
    0x806567CB,
    0x196C3671,
    0x6E6B06E7,
    0xFED41B76,
    0x89D32BE0,
    0x10DA7A5A,
    0x67DD4ACC,
    0xF9B9DF6F,
    0x8EBEEFF9,
    0x17B7BE43,
    0x60B08ED5,
    0xD6D6A3E8,
    0xA1D1937E,
    0x38D8C2C4,
    0x4FDFF252,
    0xD1BB67F1,
    0xA6BC5767,
    0x3FB506DD,
    0x48B2364B,
    0xD80D2BDA,
    0xAF0A1B4C,
    0x36034AF6,
    0x41047A60,
    0xDF60EFC3,
    0xA867DF55,
    0x316E8EEF,
    0x4669BE79,
    0xCB61B38C,
    0xBC66831A,
    0x256FD2A0,
    0x5268E236,
    0xCC0C7795,
    0xBB0B4703,
    0x220216B9,
    0x5505262F,
    0xC5BA3BBE,
    0xB2BD0B28,
    0x2BB45A92,
    0x5CB36A04,
    0xC2D7FFA7,
    0xB5D0CF31,
    0x2CD99E8B,
    0x5BDEAE1D,
    0x9B64C2B0,
    0xEC63F226,
    0x756AA39C,
    0x026D930A,
    0x9C0906A9,
    0xEB0E363F,
    0x72076785,
    0x05005713,
    0x95BF4A82,
    0xE2B87A14,
    0x7BB12BAE,
    0x0CB61B38,
    0x92D28E9B,
    0xE5D5BE0D,
    0x7CDCEFB7,
    0x0BDBDF21,
    0x86D3D2D4,
    0xF1D4E242,
    0x68DDB3F8,
    0x1FDA836E,
    0x81BE16CD,
    0xF6B9265B,
    0x6FB077E1,
    0x18B74777,
    0x88085AE6,
    0xFF0F6A70,
    0x66063BCA,
    0x11010B5C,
    0x8F659EFF,
    0xF862AE69,
    0x616BFFD3,
    0x166CCF45,
    0xA00AE278,
    0xD70DD2EE,
    0x4E048354,
    0x3903B3C2,
    0xA7672661,
    0xD06016F7,
    0x4969474D,
    0x3E6E77DB,
    0xAED16A4A,
    0xD9D65ADC,
    0x40DF0B66,
    0x37D83BF0,
    0xA9BCAE53,
    0xDEBB9EC5,
    0x47B2CF7F,
    0x30B5FFE9,
    0xBDBDF21C,
    0xCABAC28A,
    0x53B39330,
    0x24B4A3A6,
    0xBAD03605,
    0xCDD70693,
    0x54DE5729,
    0x23D967BF,
    0xB3667A2E,
    0xC4614AB8,
    0x5D681B02,
    0x2A6F2B94,
    0xB40BBE37,
    0xC30C8EA1,
    0x5A05DF1B,
    0x2D02EF8D,
]


class TuyaCipher:
    """Tuya cryptographic helpers."""

    def __init__(self, key: str, version: tuple[int, int]):
        """Initialize the cipher."""
        self.version = version
        self.key = key
        self.cipher = Cipher(
            algorithms.AES(key.encode("ascii")),
            modes.ECB(),  # noqa: S305
            backend=openssl_backend,
        )

    def get_prefix_size_and_validate(self, command: Message, encrypted_data: bytes) -> int:
        try:
            version = tuple(map(int, encrypted_data[:3].decode("utf8").split(".")))
        except UnicodeDecodeError:
            version = (0, 0)
        if version != self.version:
            return 0
        if version < (3, 3):
            actual_hash = encrypted_data[3:19].decode("ascii")
            expected_hash = self.hash(encrypted_data[19:])
            if actual_hash != expected_hash:
                return 0
            return 19
        elif command in (Message.SET_COMMAND, Message.GRATUITOUS_UPDATE):
            _, sequence, __, ___ = struct.unpack_from(">IIIH", encrypted_data, 3)
            return 15
        return 0

    def decrypt(self, command: Message, data: bytes) -> bytes:
        prefix_size = self.get_prefix_size_and_validate(command, data)
        data = data[prefix_size:]
        decryptor = self.cipher.decryptor()
        if self.version < (3, 3):
            data = base64.b64decode(data)
        decrypted_data = decryptor.update(data)
        decrypted_data += decryptor.finalize()
        unpadder = PKCS7(128).unpadder()
        unpadded_data = unpadder.update(decrypted_data)
        unpadded_data += unpadder.finalize()

        return unpadded_data

    def encrypt(self, command: int, data: bytes) -> bytes:
        encrypted_data = b""
        if data:
            padder = PKCS7(128).padder()
            padded_data = padder.update(data)
            padded_data += padder.finalize()
            encryptor = self.cipher.encryptor()
            encrypted_data = encryptor.update(padded_data)
            encrypted_data += encryptor.finalize()

        prefix = ".".join(map(str, self.version)).encode("utf8")
        if self.version < (3, 3):
            payload = base64.b64encode(encrypted_data)
            payload_hash = self.hash(payload)
            prefix += payload_hash.encode("utf8")
        else:
            payload = encrypted_data
            if command in (Message.SET_COMMAND, Message.GRATUITOUS_UPDATE):
                prefix += b"\x00" * 12
            else:
                prefix = b""

        return prefix + payload

    def hash(self, data: bytes) -> str:
        digest = Hash(MD5(), backend=openssl_backend)  # noqa: S303
        to_hash = "data={}||lpv={}||{}".format(
            data.decode("ascii"), ".".join(map(str, self.version)), self.key
        )
        digest.update(to_hash.encode("utf8"))
        intermediate = digest.finalize().hex()
        return intermediate[8:24]


def crc(data: bytes) -> int:
    """Calculate the Tuya-flavored CRC of some data."""
    c = 0xFFFFFFFF
    for b in data:
        c = (c >> 8) ^ CRC_32_TABLE[(c ^ b) & 255]

    return c ^ 0xFFFFFFFF


class Message:
    PING_COMMAND = 0x09
    GET_COMMAND = 0x0A
    SET_COMMAND = 0x07
    GRATUITOUS_UPDATE = 0x08

    def __init__(
        self,
        command: int,
        payload: dict[str, Any] | bytes | None = None,
        sequence: int | None = None,
        encrypt_for: TuyaDevice | None = None,
    ):
        if payload is None:
            payload = b""
        self.payload = payload
        self.command = command
        if sequence is None:
            # Use millisecond process time as the sequence number. Not ideal,
            # but good for one month's continuous connection time though.
            sequence = int(time.perf_counter() * 1000) & 0xFFFFFFFF
        self.sequence = sequence
        self.encrypt = False
        self.device = None
        if encrypt_for is not None:
            self.device = encrypt_for
            self.encrypt = True

    def __repr__(self) -> str:
        return "{}({}, {!r}, {!r}, {})".format(
            self.__class__.__name__,
            hex(self.command),
            self.payload,
            self.sequence,
            f"<Device {self.device}>" if self.device else None,
        )

    def hex(self) -> str:
        return self.to_bytes().hex()

    def to_bytes(self) -> bytes:
        if isinstance(self.payload, dict):
            payload_data = json.dumps(self.payload, separators=(",", ":")).encode("utf-8")
        elif not isinstance(self.payload, bytes):
            payload_data = self.payload.encode("utf8")
        else:
            payload_data = self.payload

        if self.encrypt:
            assert self.device
            payload_data = self.device.cipher.encrypt(self.command, payload_data)

        payload_size = len(payload_data) + struct.calcsize(MESSAGE_SUFFIX_FORMAT)

        header = struct.pack(
            MESSAGE_PREFIX_FORMAT,
            MAGIC_PREFIX,
            self.sequence,
            self.command,
            payload_size,
        )
        if self.device and self.device.version >= (3, 3):
            checksum = crc(header + payload_data)
        else:
            checksum = crc(payload_data)
        footer = struct.pack(MESSAGE_SUFFIX_FORMAT, checksum, MAGIC_SUFFIX)
        return header + payload_data + footer

    __bytes__ = bytes

    @classmethod
    def from_bytes(cls, data: bytes, cipher: TuyaCipher) -> Self:
        try:
            prefix, sequence, command, payload_size = struct.unpack_from(
                MESSAGE_PREFIX_FORMAT, data
            )
        except struct.error as e:
            raise InvalidMessage("Invalid message header format.") from e
        if prefix != MAGIC_PREFIX:
            raise InvalidMessage("Magic prefix missing from message.")

        # check for an optional return code
        header_size = struct.calcsize(MESSAGE_PREFIX_FORMAT)
        try:
            (return_code,) = struct.unpack_from(">I", data, header_size)
        except struct.error as e:
            raise InvalidMessage("Unable to unpack return code.") from e
        if return_code >> 8:
            payload_data = data[
                header_size : header_size + payload_size - struct.calcsize(MESSAGE_SUFFIX_FORMAT)
            ]
            return_code = None
        else:
            payload_data = data[
                header_size + struct.calcsize(">I") : header_size
                + payload_size
                - struct.calcsize(MESSAGE_SUFFIX_FORMAT)
            ]

        try:
            expected_crc, suffix = struct.unpack_from(
                MESSAGE_SUFFIX_FORMAT,
                data,
                header_size + payload_size - struct.calcsize(MESSAGE_SUFFIX_FORMAT),
            )
        except struct.error as e:
            raise InvalidMessage("Invalid message suffix format.") from e
        if suffix != MAGIC_SUFFIX:
            raise InvalidMessage("Magic suffix missing from message")

        actual_crc = crc(
            data[: header_size + payload_size - struct.calcsize(MESSAGE_SUFFIX_FORMAT)]
        )
        if expected_crc != actual_crc:
            raise InvalidMessage("CRC check failed")

        payload = None
        if payload_data:
            with contextlib.suppress(ValueError):
                payload_data = cipher.decrypt(command, payload_data)
            try:
                payload_text = payload_data.decode("utf8")
            except UnicodeDecodeError as e:
                _LOGGER.debug(payload_data.hex())
                _LOGGER.error(e)
                raise MessageDecodeFailed from e
            try:
                payload = json.loads(payload_text)
            except json.decoder.JSONDecodeError as e:
                # data may be encrypted
                _LOGGER.debug(payload_data.hex())
                _LOGGER.error(e)
                raise MessageDecodeFailed from e

        return cls(command, payload, sequence)


class TuyaDevice:
    """Represents a generic Tuya device."""

    PING_INTERVAL = 10

    def __init__(
        self,
        unique_id: str,
        host: str,
        local_key: str,
        port: int = 6668,
        gateway_id: str | None = None,
        version: tuple[int, int] = (3, 3),
        timeout: int = 10,
    ):
        """Initialize the device."""
        self.unique_id = unique_id
        self.host = host
        self.port = port
        if not gateway_id:
            gateway_id = self.unique_id
        self.gateway_id = gateway_id
        self.version = version
        self.timeout = timeout
        self.last_pong = 0.0

        if len(local_key) != 16:
            raise InvalidKey("Local key should be a 16-character string")

        self.cipher = TuyaCipher(local_key, self.version)
        self.writer: asyncio.StreamWriter | None = None
        self._futures: dict[int, asyncio.Future[Message]] = {}
        self._dps: dict[str, str | int | float] = {}
        self._connected = False
        self._connecting_lock = asyncio.Lock()
        self._availability_callbacks: set[Callable[[bool], None]] = set()
        self._state_callbacks: set[Callable[[Any], None]] = set()

    def __str__(self) -> str:
        return f"{self.unique_id} ({self.host}:{self.port})"

    def async_add_availability_callback(
        self, callback: Callable[[Any], None]
    ) -> Callable[[], None]:
        self._availability_callbacks.add(callback)

        def _() -> None:
            self._availability_callbacks.discard(callback)

        return _

    def async_add_state_callback(self, callback: Callable[[Any], None]) -> Callable[[], None]:
        self._state_callbacks.add(callback)

        def _() -> None:
            self._state_callbacks.discard(callback)

        return _

    async def async_connect(self) -> None:
        async with self._connecting_lock:
            if self._connected:
                return

            _LOGGER.debug("Starting connection to %s:%s", self.host, self.port)

            try:
                self.reader, self.writer = await asyncio.wait_for(
                    asyncio.open_connection(host=self.host, port=self.port), self.timeout
                )
            except ConnectionRefusedError as e:
                raise ConnectionFailed(
                    f"Could not start connection to {self.host}:{self.port} - "
                    "device is offline or ip is incorrect"
                ) from e
            except TimeoutError as e:
                raise ConnectionTimeoutException("Connection timed out") from e

            self._connected = True
            for callback in self._availability_callbacks:
                callback(True)

        await self._async_ping()

    def _async_disconnect(self) -> None:
        _LOGGER.debug(f"Disconnected from {self}")
        self._connected = False
        for callback in self._availability_callbacks:
            callback(False)
        self.last_pong = 0
        if self.writer is not None:
            self.writer.close()
            self.writer = None

    async def async_poll(self) -> None:
        payload = {"gwId": self.gateway_id, "devId": self.unique_id}
        maybe_self = None if self.version < (3, 3) else self
        message = Message(Message.GET_COMMAND, payload, encrypt_for=maybe_self)
        await self.async_call(message)

    async def async_set(self, dps: dict[str, Any]) -> Message:
        t = int(time.time())
        payload = {"devId": self.unique_id, "uid": "", "t": t, "dps": dps}
        message = Message(Message.SET_COMMAND, payload, encrypt_for=self)
        return await self.async_call(message)

    async def _async_ping(self) -> None:
        self.last_ping = time.time()
        maybe_self = None if self.version < (3, 3) else self
        message = Message(Message.PING_COMMAND, sequence=0, encrypt_for=maybe_self)
        await self.async_send(message)

    async def async_update_state(self, state_message: Message) -> None:
        _LOGGER.debug("Received updated state %s: %s", self, state_message)
        assert isinstance(state_message.payload, dict)
        self._dps.update(state_message.payload["dps"])
        self.state = self._handle_state_update(self._dps)
        _LOGGER.debug("New vacuum state %s: %s", self, self.state)
        for callback in self._state_callbacks:
            callback(self.state)

    def _handle_state_update(self, dps: dict[str, Any]) -> Any:
        raise NotImplementedError

    async def _async_read_message(self) -> Message | None:
        try:
            response_data = await self.reader.readuntil(MAGIC_SUFFIX_BYTES)
        except OSError as e:
            _LOGGER.error(f"Connection to {self} failed: {e}")
            self._async_disconnect()
            return None

        try:
            message = Message.from_bytes(response_data, self.cipher)

        except InvalidMessage as e:
            _LOGGER.error(f"Invalid message from {self}: {e}")
            self._async_disconnect()
            return None

        except MessageDecodeFailed:
            _LOGGER.error(f"Failed to decrypt message from {self}")
            self._async_disconnect()
            return None

        _LOGGER.debug(f"Received message from {self}: {message}")
        return message

    async def async_process_messages(self) -> None:
        while True:
            try:
                await self.async_connect()
            except ConnectionFailed:
                _LOGGER.debug("Could not connect to %s:%s", self.host, self.port)
                await asyncio.sleep(10)

            sleep_fut = asyncio.create_task(asyncio.sleep(self.PING_INTERVAL))
            message_fut = asyncio.create_task(self._async_read_message())

            try:
                while self._connected:
                    (done, pending) = await asyncio.wait(
                        [message_fut, sleep_fut], return_when=asyncio.FIRST_COMPLETED
                    )

                    if message_fut in done:
                        message = await message_fut

                        if not message:
                            break

                        if message.command in (Message.GET_COMMAND, Message.GRATUITOUS_UPDATE):
                            await self.async_update_state(message)

                        elif message.command == Message.PING_COMMAND:
                            self.last_pong = time.time()

                        if future := self._futures.pop(message.sequence, None):
                            future.set_result(message)

                        message_fut = asyncio.create_task(self._async_read_message())

                    if sleep_fut in done:
                        if self.last_pong < self.last_ping:
                            self._async_disconnect()
                            break
                        await sleep_fut
                        await self._async_ping()
                        sleep_fut = asyncio.create_task(asyncio.sleep(self.PING_INTERVAL))
            finally:
                if not message_fut.done():
                    message_fut.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await message_fut

                if not sleep_fut.done():
                    sleep_fut.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await message_fut

    async def async_send(self, message: Message, retries: int = 4) -> None:
        await self.async_connect()
        _LOGGER.debug(f"Sending to {self}: {message}")
        try:
            assert self.writer
            self.writer.write(message.to_bytes())
        except (TimeoutError, OSError) as e:
            if retries == 0:
                raise ConnectionException(f"Failed to send data to {self}") from e
            await self.async_connect()
            await self.async_send(message, retries=retries - 1)

    async def async_call(self, message: Message, retries: int = 4) -> Message:
        fut = self._futures[message.sequence] = asyncio.Future()

        try:
            await self.async_send(message, retries=retries)
        except ConnectionException:
            self._futures.pop(message.sequence, None)

        try:
            return await asyncio.wait_for(fut, 10)
        except TimeoutError:
            self._futures.pop(message.sequence, None)
            raise



"""
For obtaining a list of keys for controlling eufy devices.

This is only needed once, to get the initial key.

See https://github.com/markbajaj/eufy-device-id-python/blob/main/eufy_local_id_grabber/crypto.py
"""

import json
import logging
from dataclasses import dataclass
from typing import Final

import aiohttp

from .countries import (
    phone_code_for_country,
    phone_code_for_region,
    region_for_country,
    region_for_phone_code,
)
from .exceptions import AuthenticationFailed, ConnectionFailed
from .tuya_cloud import TuyaAPISession

_LOGGER = logging.getLogger(__name__)

EUFY_HEADERS: Final[dict[str, str]] = {
    "User-Agent": "EufyHome-Android-2.4.0",
    "timezone": "Europe/London",
    "category": "Home",
    "token": "",
    "uid": "",
    "openudid": "sdk_gphone64_arm64",
    "clientType": "2",
    "language": "en",
    "country": "US",
    "Accept-Encoding": "gzip",
}


@dataclass
class VacuumCloudDiscovery:
    id: str
    name: str
    description: str
    model: str
    mac: str
    access_token: str


async def get_cloud_vacuums(
    session: aiohttp.ClientSession, username: str, password: str
) -> dict[str, VacuumCloudDiscovery]:
    """Login to Eufy and get the vacuum details."""
    user_info_resp = await session.post(
        "https://home-api.eufylife.com/v1/user/email/login",
        headers=EUFY_HEADERS,
        json={
            "client_Secret": "GQCpr9dSp3uQpsOMgJ4xQ",
            "client_id": "eufyhome-app",
            "email": username,
            "password": password,
        },
    )

    if user_info_resp.status != 200:
        raise ConnectionFailed

    user_info = await user_info_resp.json()

    if user_info["res_code"] != 1:
        raise AuthenticationFailed

    user_id = user_info["user_info"]["id"]
    host = user_info["user_info"]["request_host"]
    access_token = user_info["access_token"]

    device_info_resp = await session.get(
        host + "/v1/device/list/devices-and-groups",
        headers={**EUFY_HEADERS, "token": access_token, "id": user_id},
    )
    device_info = await device_info_resp.json()

    user_settings_resp = await session.get(
        host + "/v1/user/setting",
        headers={**EUFY_HEADERS, "token": access_token, "id": user_id},
    )
    user_settings = await user_settings_resp.json()

    if (
        "tuya_home" in user_settings["setting"]["home_setting"]
        and "tuya_region_code" in user_settings["setting"]["home_setting"]["tuya_home"]
    ):
        region = user_settings["setting"]["home_setting"]["tuya_home"]["tuya_region_code"]
        if user_info["user_info"]["phone_code"]:
            country_code = user_info["user_info"]["phone_code"]
        else:
            country_code = phone_code_for_region(region)
    elif user_info["user_info"]["phone_code"]:
        region = region_for_phone_code(user_info["user_info"]["phone_code"])
        country_code = user_info["user_info"]["phone_code"]
    elif user_info["user_info"]["country"]:
        region = region_for_country(user_info["user_info"]["country"])
        country_code = phone_code_for_country(user_info["user_info"]["country"])
    else:
        region = "EU"
        country_code = "44"

    time_zone = user_info["user_info"]["timezone"]

    tuya_client = TuyaAPISession(
        session,
        username="eh-" + user_id,
        region=region,
        timezone=time_zone,
        phone_code=country_code,
    )

    vacs = {}
    for item in device_info["items"]:
        if item["device"]["product"]["appliance"] != "Cleaning":
            continue

        try:
            device = await tuya_client.get_device(item["device"]["id"])
        except Exception:
            _LOGGER.debug(
                "Skipping vacuum %s: found on Eufy but not on Tuya. Eufy details:",
                item["device"]["id"],
            )
            _LOGGER.debug(json.dumps(item["device"], indent=2))
            continue

        vacs[item["device"]["id"]] = VacuumCloudDiscovery(
            id=item["device"]["id"],
            name=item["device"]["alias_name"],
            description=item["device"]["name"],
            model=item["device"]["product"]["product_code"],
            mac=item["device"]["wifi"]["mac"],
            access_token=device["localKey"],
        )

    return vacs



"""
For obtaining a list of keys for controlling eufy devices.

This is only needed once, to get the initial key.

See https://github.com/markbajaj/eufy-device-id-python/blob/main/eufy_local_id_grabber/crypto.py
"""

import hmac
import json
import math
import random
import string
import time
import uuid
from hashlib import md5, sha256
from typing import Any

import aiohttp
from cryptography.hazmat.backends.openssl import backend as openssl_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

EUFY_HMAC_KEY = b"A_cepev5pfnhua4dkqkdpmnrdxx378mpjr_s8x78u7xwymasd9kqa7a73pjhxqsedaj"


def unpadded_rsa(key_exponent: int, key_n: int, plaintext: bytes) -> bytes:
    keylength = math.ceil(key_n.bit_length() / 8)
    input_nr = int.from_bytes(plaintext, byteorder="big")
    crypted_nr = pow(input_nr, key_exponent, key_n)
    return crypted_nr.to_bytes(keylength, byteorder="big")


def shuffled_md5(value: str) -> str:
    _hash = md5(value.encode("utf-8")).hexdigest()
    return _hash[8:16] + _hash[0:8] + _hash[24:32] + _hash[16:24]


TUYA_PASSWORD_INNER_CIPHER = Cipher(
    algorithms.AES(
        bytearray([36, 78, 109, 138, 86, 172, 135, 145, 36, 67, 45, 139, 108, 188, 162, 196])
    ),
    modes.CBC(bytearray([119, 36, 86, 242, 167, 102, 76, 243, 57, 44, 53, 151, 233, 62, 87, 71])),
    backend=openssl_backend,
)

DEFAULT_TUYA_HEADERS = {"User-Agent": "TY-UA=APP/Android/2.4.0/SDK/null"}

SIGNATURE_RELEVANT_PARAMETERS = {
    "a",
    "v",
    "lat",
    "lon",
    "lang",
    "deviceId",
    "appVersion",
    "ttid",
    "isH5",
    "h5Token",
    "os",
    "clientId",
    "postData",
    "time",
    "requestId",
    "et",
    "n4h5",
    "sid",
    "sp",
}

DEFAULT_TUYA_QUERY_PARAMS = {
    "appVersion": "2.4.0",
    "deviceId": "",
    "platform": "sdk_gphone64_arm64",
    "clientId": "yx5v9uc3ef9wg3v9atje",
    "lang": "en",
    "osSystem": "12",
    "os": "Android",
    "timeZoneId": "Europe/London",
    "ttid": "android",
    "et": "0.0.1",
    "sdkVersion": "3.0.8cAnker",
}


class TuyaAPISession:
    session_id = None

    def __init__(
        self,
        session: aiohttp.ClientSession,
        username: str,
        region: str,
        timezone: str,
        phone_code: str,
    ):
        self.session = session
        self.headers = DEFAULT_TUYA_HEADERS.copy()
        self.default_query_params = DEFAULT_TUYA_QUERY_PARAMS.copy()
        self.default_query_params["deviceId"] = self.generate_new_device_id()
        self.default_query_params["timeZoneId"] = timezone
        self.username = username
        self.country_code = phone_code
        self.base_url = {
            "AZ": "https://a1.tuyaus.com",
            "AY": "https://a1.tuyacn.com",
            "IN": "https://a1.tuyain.com",
            "EU": "https://a1.tuyaeu.com",
        }.get(region, "https://a1.tuyaeu.com")

    @staticmethod
    def generate_new_device_id() -> str:
        expected_length = 44
        base64_characters = string.ascii_letters + string.digits
        device_id_dependent_part = "8534c8ec0ed0"
        return device_id_dependent_part + "".join(
            random.choice(base64_characters)
            for _ in range(expected_length - len(device_id_dependent_part))
        )

    @staticmethod
    def get_signature(query_params: dict[str, str], encoded_post_data: str) -> str:
        query_params = query_params.copy()
        if encoded_post_data:
            query_params["postData"] = encoded_post_data
        sorted_pairs = sorted(query_params.items())
        filtered_pairs = filter(
            lambda p: p[0] and p[0] in SIGNATURE_RELEVANT_PARAMETERS, sorted_pairs
        )
        mapped_pairs = (
            p[0] + "=" + (shuffled_md5(p[1]) if p[0] == "postData" else p[1])
            for p in filtered_pairs
        )
        message = "||".join(mapped_pairs)
        return hmac.HMAC(
            key=EUFY_HMAC_KEY, msg=message.encode("utf-8"), digestmod=sha256
        ).hexdigest()

    async def _request(
        self,
        action: str,
        version: str = "1.0",
        data: dict[str, str] | None = None,
        query_params: dict[str, str] | None = None,
        _requires_session: bool = True,
    ) -> Any:
        if not self.session_id and _requires_session:
            await self.acquire_session()

        current_time = time.time()
        request_id = uuid.uuid4()
        extra_query_params = {
            "time": str(int(current_time)),
            "requestId": str(request_id),
            "a": action,
            "v": version,
            **(query_params or {}),
        }
        query_params = {**self.default_query_params, **extra_query_params}
        encoded_post_data = json.dumps(data, separators=(",", ":")) if data else ""
        resp = await self.session.post(
            self.base_url + "/api.json",
            headers=self.headers,
            params={
                **query_params,
                "sign": self.get_signature(query_params, encoded_post_data),
            },
            data={"postData": encoded_post_data} if encoded_post_data else None,
            raise_for_status=True,
        )
        resp_json = await resp.json()
        if "result" not in resp_json:
            raise Exception(
                f"No 'result' key in the response - the entire response is {resp_json}."
            )
        return resp_json["result"]

    async def request_token(self, username: str, country_code: str) -> Any:
        return await self._request(
            action="tuya.m.user.uid.token.create",
            data={"uid": username, "countryCode": country_code},
            _requires_session=False,
        )

    def determine_password(self, username: str) -> str:
        new_uid = username
        padded_size = 16 * math.ceil(len(new_uid) / 16)
        password_uid = new_uid.zfill(padded_size)
        encryptor = TUYA_PASSWORD_INNER_CIPHER.encryptor()
        encrypted_uid = encryptor.update(password_uid.encode("utf8"))
        encrypted_uid += encryptor.finalize()
        return md5(encrypted_uid.hex().upper().encode("utf-8")).hexdigest()

    async def request_session(self, username: str, country_code: str) -> Any:
        password = self.determine_password(username)
        token_response = await self.request_token(username, country_code)
        encrypted_password = unpadded_rsa(
            key_exponent=int(token_response["exponent"]),
            key_n=int(token_response["publicKey"]),
            plaintext=password.encode("utf-8"),
        )
        data = {
            "uid": username,
            "createGroup": True,
            "ifencrypt": 1,
            "passwd": encrypted_password.hex(),
            "countryCode": country_code,
            "options": '{"group": 1}',
            "token": token_response["token"],
        }
        return await self._request(
            action="tuya.m.user.uid.password.login.reg",
            data=data,
            _requires_session=False,
        )

    async def acquire_session(self) -> None:
        session_response = await self.request_session(self.username, self.country_code)
        self.session_id = self.default_query_params["sid"] = session_response["sid"]
        self.base_url = session_response["domain"]["mobileApiUrl"]

    def get_device(self, device_id: str) -> Any:
        return self._request(action="tuya.m.device.get", version="1.0", data={"devId": device_id})



from .const import CleanSpeed
from .device import BinarySensor, Sensor, Switch, VacuumDevice, VacuumState
from .eufy_cloud import VacuumCloudDiscovery, get_cloud_vacuums
from .exceptions import (
    AuthenticationFailed,
    ConnectionException,
    ConnectionFailed,
    ConnectionTimeoutException,
    EufyCleanException,
    InvalidKey,
    InvalidMessage,
    MessageDecodeFailed,
)
from .metadata import VACUUM_INFO, VACUUM_MODEL_NAME_TO_ID, VacuumInfo

__all__ = [
    "EufyCleanException",
    "ConnectionFailed",
    "AuthenticationFailed",
    "ConnectionException",
    "ConnectionTimeoutException",
    "InvalidKey",
    "InvalidMessage",
    "MessageDecodeFailed",
    "get_cloud_vacuums",
    "VacuumCloudDiscovery",
    "VacuumDevice",
    "VacuumState",
    "Switch",
    "BinarySensor",
    "Sensor",
    "CleanSpeed",
    "VACUUM_INFO",
    "VacuumInfo",
    "VACUUM_MODEL_NAME_TO_ID",
]



# Copyright 2019 Richard Mitchell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
from dataclasses import dataclass
from enum import StrEnum

from aioeufyclean.metadata import VACUUM_INFO

from .const import CleanSpeed, WorkMode
from .tuya import TuyaDevice


class ErrorCode(StrEnum):
    NO_ERROR = "no_error"
    WHEEL_STUCK = "Wheel_stuck"
    R_BRUSH_STUCK = "R_brush_stuck"
    CRASH_BAR_STUCK = "Crash_bar_stuck"
    SENSOR_DIRTY = "sensor_dirty"
    NOT_ENOUGH_POWER = "N_enough_pow"
    STUCK_5_MIN = "Stuck_5_min"
    FAN_STUCK = "Fan_stuck"
    S_BRUSH_STUCK = "S_brush_stuck"


class State(StrEnum):
    DOCKED = "docked"
    CLEANING = "cleaning"
    RETURNING = "returning"
    ERROR = "error"
    PAUSED = "paused"
    ON = "on"
    OFF = "off"
    IDLE = "idle"


class BinarySensor(StrEnum):
    pass


class Sensor(StrEnum):
    BATTERY = "battery"
    FILTER_LIFE = "filter_life"
    SIDE_BRUSH_LIFE = "side_brush_life"
    ROLLING_BRUSH_LIFE = "rolling_brush_life"
    SENSOR_CLEAN_LIFE = "sensor_clean_life"


class Switch(StrEnum):
    BOOST_IQ = "boost_iq"


@dataclass
class VacuumState:
    state: State
    clean_speed: CleanSpeed
    sensors: dict[Sensor, str | int | float]
    binary_sensors: dict[BinarySensor, bool]
    switches: dict[Switch, bool]


class VacuumDevice(TuyaDevice):
    """Represents a generic Eufy Robovac."""

    def __init__(
        self,
        unique_id: str,
        host: str,
        local_key: str,
        model_id: str,
        port: int = 6668,
        gateway_id: str | None = None,
        version: tuple[int, int] = (3, 3),
        timeout: int = 10,
    ):
        self.model_id = model_id
        self.device_info = VACUUM_INFO[model_id]

        super().__init__(unique_id, host, local_key, port, gateway_id, version, timeout)

    def _handle_state_update(self, payload: dict[str, str | int | float]) -> VacuumState:
        if payload.get(self.device_info.error_code) != 0:
            state = State.ERROR
        elif payload.get(self.device_info.power) == "1" or payload.get(
            self.device_info.work_status
        ) in (
            "Charging",
            "completed",
        ):
            state = State.DOCKED
        elif payload.get(self.device_info.work_status) in ("Recharge",):
            state = State.RETURNING
        elif payload.get(self.device_info.work_status) in ("Sleeping", "standby"):
            state = State.IDLE
        else:
            state = State.CLEANING

        clean_speed = CleanSpeed(self.device_info.clean_speed)

        vacuum_state = VacuumState(
            state=state,
            clean_speed=clean_speed,
            sensors={},
            binary_sensors={},
            switches={},
        )

        if self.device_info.battery_level and self.device_info.battery_level in payload:
            if battery_level := payload.get(self.device_info.battery_level):
                vacuum_state.sensors[Sensor.BATTERY] = int(battery_level)

        if self.device_info.consumable:
            if consumable_json := payload.get(self.device_info.consumable):
                if (
                    duration := json.loads(base64.b64decode(str(consumable_json)))
                    .get("consumable", {})
                    .get("duration", {})
                ):
                    # TODO: What are SP, TR and BatteryStatus?
                    if "FM" in duration:
                        vacuum_state.sensors[Sensor.FILTER_LIFE] = duration["FM"]
                    if "RB" in duration:
                        vacuum_state.sensors[Sensor.ROLLING_BRUSH_LIFE] = duration["RB"]
                    if "SB" in duration:
                        vacuum_state.sensors[Sensor.SIDE_BRUSH_LIFE] = duration["SB"]
                    if "SS" in duration:
                        vacuum_state.sensors[Sensor.SENSOR_CLEAN_LIFE] = duration["SS"]

        if self.device_info.boost_iq:
            if boost_iq := payload.get(self.device_info.boost_iq):
                vacuum_state.switches[Switch.BOOST_IQ] = bool(boost_iq)

        return vacuum_state

    async def async_start(self) -> None:
        await self.async_set({self.device_info.work_mode: str(WorkMode.AUTO)})

    async def async_pause(self) -> None:
        await self.async_set({self.device_info.play_pause: False})

    async def async_stop(self) -> None:
        await self.async_set({self.device_info.play_pause: False})

    async def async_return_to_base(self) -> None:
        await self.async_set({self.device_info.go_home: True})

    async def async_locate(self) -> None:
        await self.async_set({self.device_info.find_robot: True})

    async def async_set_fan_speed(self, clean_speed: CleanSpeed) -> None:
        await self.async_set({self.device_info.clean_speed: str(clean_speed)})

    async def async_clean_spot(self) -> None:
        await self.async_set({self.device_info.work_mode: WorkMode.SPOT})

    async def async_set_switch(self, switch: Switch, value: bool) -> None:
        if self.device_info.boost_iq and switch == Switch.BOOST_IQ:
            await self.async_set({self.device_info.boost_iq: value})



class EufyCleanException(Exception):
    pass


class ConnectionFailed(EufyCleanException):
    pass


class AuthenticationFailed(EufyCleanException):
    pass


class TuyaException(EufyCleanException):
    """Base for Tuya exceptions."""


class InvalidKey(TuyaException):
    """The local key is invalid."""


class InvalidMessage(TuyaException):
    """The message received is invalid."""


class MessageDecodeFailed(TuyaException):
    """The message received cannot be decoded as JSON."""


class ConnectionException(TuyaException):
    """The socket connection failed."""


class ConnectionTimeoutException(ConnectionException):
    """The socket connection timed out."""


class RequestResponseCommandMismatch(TuyaException):
    """The command in the response didn't match the one from the request."""



_COUNTRIES = [
    {"country": "AE", "phone_code": "971", "region": "EU"},
    {"country": "AF", "phone_code": "93", "region": "EU"},
    {"country": "AL", "phone_code": "355", "region": "EU"},
    {"country": "AM", "phone_code": "374", "region": "EU"},
    {"country": "AO", "phone_code": "244", "region": "EU"},
    {"country": "AR", "phone_code": "54", "region": "AZ"},
    {"country": "AT", "phone_code": "43", "region": "EU"},
    {"country": "AU", "phone_code": "61", "region": "AZ"},
    {"country": "AZ", "phone_code": "994", "region": "EU"},
    {"country": "BA", "phone_code": "387", "region": "EU"},
    {"country": "BD", "phone_code": "880", "region": "EU"},
    {"country": "BE", "phone_code": "32", "region": "EU"},
    {"country": "BF", "phone_code": "226", "region": "EU"},
    {"country": "BG", "phone_code": "359", "region": "EU"},
    {"country": "BH", "phone_code": "973", "region": "EU"},
    {"country": "BI", "phone_code": "257", "region": "EU"},
    {"country": "BJ", "phone_code": "229", "region": "EU"},
    {"country": "BN", "phone_code": "673", "region": "EU"},
    {"country": "BO", "phone_code": "591", "region": "AZ"},
    {"country": "BR", "phone_code": "55", "region": "AZ"},
    {"country": "BT", "phone_code": "975", "region": "EU"},
    {"country": "BW", "phone_code": "267", "region": "EU"},
    {"country": "BY", "phone_code": "375", "region": "EU"},
    {"country": "BZ", "phone_code": "501", "region": "EU"},
    {"country": "CA", "phone_code": "1", "region": "AZ"},
    {"country": "CD", "phone_code": "243", "region": "EU"},
    {"country": "CF", "phone_code": "236", "region": "EU"},
    {"country": "CG", "phone_code": "242", "region": "EU"},
    {"country": "CH", "phone_code": "41", "region": "EU"},
    {"country": "CI", "phone_code": "225", "region": "EU"},
    {"country": "CL", "phone_code": "56", "region": "AZ"},
    {"country": "CM", "phone_code": "237", "region": "EU"},
    {"country": "CN", "phone_code": "86", "region": "AY"},
    {"country": "CO", "phone_code": "57", "region": "AZ"},
    {"country": "CR", "phone_code": "506", "region": "EU"},
    {"country": "CV", "phone_code": "238", "region": "EU"},
    {"country": "CY", "phone_code": "357", "region": "EU"},
    {"country": "CZ", "phone_code": "420", "region": "EU"},
    {"country": "DE", "phone_code": "49", "region": "EU"},
    {"country": "DJ", "phone_code": "253", "region": "EU"},
    {"country": "DK", "phone_code": "45", "region": "EU"},
    {"country": "DO", "phone_code": "1809", "region": "EU"},
    {"country": "DO", "phone_code": "1829", "region": "EU"},
    {"country": "DO", "phone_code": "1849", "region": "EU"},
    {"country": "DZ", "phone_code": "213", "region": "EU"},
    {"country": "EC", "phone_code": "593", "region": "AZ"},
    {"country": "EE", "phone_code": "372", "region": "EU"},
    {"country": "EG", "phone_code": "20", "region": "EU"},
    {"country": "ER", "phone_code": "291", "region": "EU"},
    {"country": "ES", "phone_code": "34", "region": "EU"},
    {"country": "ET", "phone_code": "251", "region": "EU"},
    {"country": "FI", "phone_code": "358", "region": "EU"},
    {"country": "FJ", "phone_code": "679", "region": "EU"},
    {"country": "FR", "phone_code": "33", "region": "EU"},
    {"country": "GA", "phone_code": "241", "region": "EU"},
    {"country": "GB", "phone_code": "44", "region": "EU"},
    {"country": "GE", "phone_code": "995", "region": "EU"},
    {"country": "GH", "phone_code": "233", "region": "EU"},
    {"country": "GL", "phone_code": "299", "region": "EU"},
    {"country": "GM", "phone_code": "220", "region": "EU"},
    {"country": "GN", "phone_code": "224", "region": "EU"},
    {"country": "GQ", "phone_code": "240", "region": "EU"},
    {"country": "GR", "phone_code": "30", "region": "EU"},
    {"country": "GT", "phone_code": "502", "region": "AZ"},
    {"country": "GY", "phone_code": "592", "region": "EU"},
    {"country": "HK", "phone_code": "852", "region": "AZ"},
    {"country": "HN", "phone_code": "504", "region": "EU"},
    {"country": "HR", "phone_code": "385", "region": "EU"},
    {"country": "HT", "phone_code": "509", "region": "EU"},
    {"country": "HU", "phone_code": "36", "region": "EU"},
    {"country": "ID", "phone_code": "62", "region": "AZ"},
    {"country": "IE", "phone_code": "353", "region": "EU"},
    {"country": "IL", "phone_code": "972", "region": "EU"},
    {"country": "IM", "phone_code": "44", "region": "EU"},
    {"country": "IN", "phone_code": "91", "region": "IN"},
    {"country": "IQ", "phone_code": "964", "region": "EU"},
    {"country": "IR", "phone_code": "98", "region": "EU"},
    {"country": "IS", "phone_code": "354", "region": "EU"},
    {"country": "IT", "phone_code": "39", "region": "AZ"},
    {"country": "JM", "phone_code": "1876", "region": "EU"},
    {"country": "JO", "phone_code": "962", "region": "EU"},
    {"country": "JP", "phone_code": "81", "region": "AZ"},
    {"country": "KE", "phone_code": "254", "region": "EU"},
    {"country": "KG", "phone_code": "996", "region": "EU"},
    {"country": "KH", "phone_code": "855", "region": "EU"},
    {"country": "KM", "phone_code": "269", "region": "EU"},
    {"country": "KP", "phone_code": "850", "region": "EU"},
    {"country": "KR", "phone_code": "82", "region": "AZ"},
    {"country": "KW", "phone_code": "965", "region": "EU"},
    {"country": "KY", "phone_code": "1345", "region": "EU"},
    {"country": "KZ", "phone_code": "7", "region": "EU"},
    {"country": "LA", "phone_code": "856", "region": "EU"},
    {"country": "LB", "phone_code": "961", "region": "EU"},
    {"country": "LK", "phone_code": "94", "region": "EU"},
    {"country": "LR", "phone_code": "231", "region": "EU"},
    {"country": "LS", "phone_code": "266", "region": "EU"},
    {"country": "LT", "phone_code": "370", "region": "EU"},
    {"country": "LU", "phone_code": "352", "region": "EU"},
    {"country": "LV", "phone_code": "371", "region": "EU"},
    {"country": "LY", "phone_code": "218", "region": "EU"},
    {"country": "MA", "phone_code": "212", "region": "EU"},
    {"country": "MC", "phone_code": "377", "region": "EU"},
    {"country": "MD", "phone_code": "373", "region": "EU"},
    {"country": "ME", "phone_code": "382", "region": "EU"},
    {"country": "MG", "phone_code": "261", "region": "EU"},
    {"country": "MK", "phone_code": "389", "region": "EU"},
    {"country": "ML", "phone_code": "223", "region": "EU"},
    {"country": "MM", "phone_code": "95", "region": "AZ"},
    {"country": "MN", "phone_code": "976", "region": "EU"},
    {"country": "MO", "phone_code": "853", "region": "AZ"},
    {"country": "MQ", "phone_code": "596", "region": "EU"},
    {"country": "MR", "phone_code": "222", "region": "EU"},
    {"country": "MT", "phone_code": "356", "region": "EU"},
    {"country": "MU", "phone_code": "230", "region": "EU"},
    {"country": "MV", "phone_code": "960", "region": "EU"},
    {"country": "MW", "phone_code": "265", "region": "EU"},
    {"country": "MX", "phone_code": "52", "region": "AZ"},
    {"country": "MY", "phone_code": "60", "region": "AZ"},
    {"country": "MZ", "phone_code": "258", "region": "EU"},
    {"country": "NA", "phone_code": "264", "region": "EU"},
    {"country": "NCL", "phone_code": "687", "region": "EU"},
    {"country": "NE", "phone_code": "227", "region": "EU"},
    {"country": "NG", "phone_code": "234", "region": "EU"},
    {"country": "NI", "phone_code": "505", "region": "AZ"},
    {"country": "NL", "phone_code": "31", "region": "EU"},
    {"country": "NO", "phone_code": "47", "region": "EU"},
    {"country": "NP", "phone_code": "977", "region": "EU"},
    {"country": "NZ", "phone_code": "64", "region": "AZ"},
    {"country": "OM", "phone_code": "968", "region": "EU"},
    {"country": "PA", "phone_code": "507", "region": "EU"},
    {"country": "PE", "phone_code": "51", "region": "AZ"},
    {"country": "PF", "phone_code": "689", "region": "EU"},
    {"country": "PH", "phone_code": "63", "region": "AZ"},
    {"country": "PK", "phone_code": "92", "region": "EU"},
    {"country": "PL", "phone_code": "48", "region": "EU"},
    {"country": "PR", "phone_code": "1787", "region": "AZ"},
    {"country": "PT", "phone_code": "351", "region": "EU"},
    {"country": "PY", "phone_code": "595", "region": "AZ"},
    {"country": "QA", "phone_code": "974", "region": "EU"},
    {"country": "RE", "phone_code": "262", "region": "EU"},
    {"country": "RO", "phone_code": "40", "region": "EU"},
    {"country": "RS", "phone_code": "381", "region": "EU"},
    {"country": "RU", "phone_code": "7", "region": "EU"},
    {"country": "RW", "phone_code": "250", "region": "EU"},
    {"country": "SA", "phone_code": "966", "region": "EU"},
    {"country": "SD", "phone_code": "249", "region": "EU"},
    {"country": "SE", "phone_code": "46", "region": "EU"},
    {"country": "SG", "phone_code": "65", "region": "EU"},
    {"country": "SI", "phone_code": "386", "region": "EU"},
    {"country": "SK", "phone_code": "421", "region": "EU"},
    {"country": "SL", "phone_code": "232", "region": "EU"},
    {"country": "SM", "phone_code": "378", "region": "EU"},
    {"country": "SN", "phone_code": "221", "region": "EU"},
    {"country": "SO", "phone_code": "252", "region": "EU"},
    {"country": "SR", "phone_code": "597", "region": "AZ"},
    {"country": "SV", "phone_code": "503", "region": "EU"},
    {"country": "SY", "phone_code": "963", "region": "EU"},
    {"country": "SZ", "phone_code": "268", "region": "EU"},
    {"country": "TD", "phone_code": "235", "region": "EU"},
    {"country": "TG", "phone_code": "228", "region": "EU"},
    {"country": "TH", "phone_code": "66", "region": "AZ"},
    {"country": "TJ", "phone_code": "992", "region": "EU"},
    {"country": "TM", "phone_code": "993", "region": "EU"},
    {"country": "TN", "phone_code": "216", "region": "EU"},
    {"country": "TO", "phone_code": "676", "region": "EU"},
    {"country": "TR", "phone_code": "90", "region": "EU"},
    {"country": "TT", "phone_code": "1868", "region": "EU"},
    {"country": "TW", "phone_code": "886", "region": "AZ"},
    {"country": "TZ", "phone_code": "255", "region": "EU"},
    {"country": "UA", "phone_code": "380", "region": "EU"},
    {"country": "UG", "phone_code": "256", "region": "EU"},
    {"country": "US", "phone_code": "1", "region": "AZ"},
    {"country": "UY", "phone_code": "598", "region": "AZ"},
    {"country": "UZ", "phone_code": "998", "region": "EU"},
    {"country": "VA", "phone_code": "379", "region": "EU"},
    {"country": "VE", "phone_code": "58", "region": "AZ"},
    {"country": "VG", "phone_code": "1284", "region": "EU"},
    {"country": "VI", "phone_code": "1340", "region": "EU"},
    {"country": "VN", "phone_code": "84", "region": "AZ"},
    {"country": "YE", "phone_code": "967", "region": "EU"},
    {"country": "ZA", "phone_code": "27", "region": "EU"},
    {"country": "ZM", "phone_code": "260", "region": "EU"},
    {"country": "ZR", "phone_code": "243", "region": "EU"},
    {"country": "ZW", "phone_code": "263", "region": "EU"},
]

PHONE_CODES = {v["phone_code"]: v for v in _COUNTRIES}
REGIONS = {v["region"]: v for v in _COUNTRIES}
COUNTRIES = {v["country"]: v for v in _COUNTRIES}


def region_for_country(country: str) -> str:
    return COUNTRIES.get(country, {}).get("region", "EU")


def region_for_phone_code(phone_code: str) -> str:
    return PHONE_CODES.get(phone_code, {}).get("region", "EU")


def phone_code_for_region(region: str) -> str:
    return REGIONS.get(region, {}).get("phone_code", "44")


def phone_code_for_country(country: str) -> str:
    return COUNTRIES.get(country, {}).get("phone_code", "44")



"""Metadata about eufyClean vacuums."""

from dataclasses import dataclass

from .const import CleanSpeed, WorkMode

CLEAN_SPEEDS_C = (CleanSpeed.NO_SUCTION, CleanSpeed.STANDARD, CleanSpeed.BOOST_IQ, CleanSpeed.MAX)
CLEAN_SPEEDS_G = (CleanSpeed.STANDARD, CleanSpeed.TURBO, CleanSpeed.MAX, CleanSpeed.BOOST_IQ)
CLEAN_SPEEDS_L = (CleanSpeed.QUIET, CleanSpeed.STANDARD, CleanSpeed.TURBO, CleanSpeed.MAX)
CLEAN_SPEEDS_X = (CleanSpeed.PURE, CleanSpeed.STANDARD, CleanSpeed.TURBO, CleanSpeed.MAX)


@dataclass
class VacuumInfo:
    model: str

    work_mode: str = "5"
    work_mode_list: tuple[WorkMode, ...] = (
        WorkMode.AUTO,
        WorkMode.SMALL_ROOM,
        WorkMode.SPOT,
        WorkMode.EDGE,
        WorkMode.NO_SWEEP,
    )

    clean_speed: str = "102"
    clean_speed_list: tuple[CleanSpeed, ...] = CLEAN_SPEEDS_G

    power: str = "1"
    play_pause: str = "2"
    direction: str = "3"
    work_status: str = "15"
    go_home: str = "101"
    find_robot: str = "103"
    battery_level: str = "104"
    error_code: str = "106"
    consumable: str = "116"
    boost_iq: str | None = "118"


VACUUM_INFO = {
    "T1250": VacuumInfo(model="Robovac 35C"),
    "T2080": VacuumInfo(model="Robovac S1"),
    "T2103": VacuumInfo(
        model="Robovac 11C",
        clean_speed_list=CLEAN_SPEEDS_C,
    ),
    "T2117": VacuumInfo(
        model="Robovac 35C",
        clean_speed_list=CLEAN_SPEEDS_C,
    ),
    "T2118": VacuumInfo(
        model="Robovac 30C",
        clean_speed_list=CLEAN_SPEEDS_C,
    ),
    "T2119": VacuumInfo(
        model="Robovac 11S",
        clean_speed_list=CLEAN_SPEEDS_C,
    ),
    "T2120": VacuumInfo(
        model="Robovac 15C MAX",
        clean_speed_list=CLEAN_SPEEDS_C,
    ),
    "T2123": VacuumInfo(
        model="Robovac 25C",
        clean_speed_list=CLEAN_SPEEDS_C,
    ),
    "T2128": VacuumInfo(
        model="Robovac 15C MAX",
        clean_speed_list=CLEAN_SPEEDS_C,
    ),
    "T2130": VacuumInfo(
        model="Robovac 30C MAX",
        clean_speed_list=CLEAN_SPEEDS_C,
    ),
    "T2132": VacuumInfo(
        model="Robovac 25C MAX",
        clean_speed_list=CLEAN_SPEEDS_C,
    ),
    "T2150": VacuumInfo(model="Robovac G10 Hybrid"),
    "T2181": VacuumInfo(
        model="Robovac LR30 Hybrid+",
        clean_speed_list=CLEAN_SPEEDS_L,
    ),
    "T2182": VacuumInfo(
        model="Robovac L35 Hybrid+",
        clean_speed_list=CLEAN_SPEEDS_L,
    ),
    "T2190": VacuumInfo(
        model="Robovac L70 Hybrid",
        clean_speed_list=CLEAN_SPEEDS_L,
    ),
    "T2192": VacuumInfo(model="Robovac LR20"),
    "T2193": VacuumInfo(
        model="Robovac LR30",
        clean_speed_list=CLEAN_SPEEDS_L,
    ),
    "T2194": VacuumInfo(
        model="Robovac L35 Hybrid",
        clean_speed_list=CLEAN_SPEEDS_L,
    ),
    "T2250": VacuumInfo(model="Robovac G30"),
    "T2251": VacuumInfo(model="Robovac G30 Edge"),
    "T2252": VacuumInfo(model="Robovac G30 Verge"),
    "T2253": VacuumInfo(model="Robovac G30 Hybrid"),
    "T2254": VacuumInfo(model="Robovac G35"),
    "T2255": VacuumInfo(model="Robovac G40"),
    "T2256": VacuumInfo(model="Robovac G40 Hybrid"),
    "T2257": VacuumInfo(model="Robovac G20"),
    "T2258": VacuumInfo(model="Robovac G20 Hybrid"),
    "T2259": VacuumInfo(model="Robovac G32 Pro"),
    "T2261": VacuumInfo(
        model="Robovac X8 Hybrid",
        clean_speed_list=CLEAN_SPEEDS_X,
    ),
    "T2262": VacuumInfo(
        model="Robovac X8",
        clean_speed_list=CLEAN_SPEEDS_X,
    ),
    "T2266": VacuumInfo(
        model="Robovac X8 Pro",
        clean_speed_list=CLEAN_SPEEDS_X,
    ),
    "T2267": VacuumInfo(model="RoboVac L60"),
    "T2268": VacuumInfo(model="Robovac L60 Hybrid"),
    "T2270": VacuumInfo(model="RoboVac G35+"),
    "T2272": VacuumInfo(model="Robovac G30+ SES"),
    "T2273": VacuumInfo(model="RoboVac G40 Hybrid+"),
    "T2276": VacuumInfo(model="Robovac X8 Pro SES"),
    "T2277": VacuumInfo(model="Robovac L60 SES"),
    "T2278": VacuumInfo(model="Robovac L60 Hybrid SES"),
    "T2320": VacuumInfo(
        model="Robovac X9 Pro",
        clean_speed_list=CLEAN_SPEEDS_X,
    ),
    "T2351": VacuumInfo(
        model="Robovac X10 Pro Omni",
        clean_speed_list=CLEAN_SPEEDS_X,
    ),
}

_FRIENDLY_TO_MODEL = {vi.model: k for (k, vi) in VACUUM_INFO.items()}
VACUUM_MODEL_NAME_TO_ID = {k: _FRIENDLY_TO_MODEL[k] for k in sorted(_FRIENDLY_TO_MODEL.keys())}



from enum import StrEnum


class WorkMode(StrEnum):
    AUTO = "auto"
    NO_SWEEP = "Nosweep"
    SMALL_ROOM = "SmallRoom"
    EDGE = "Edge"
    SPOT = "Spot"


class Direction(StrEnum):
    LEFT = "left"
    RIGHT = "right"
    FORWARD = "forward"
    BACKWARD = "backward"


class CleanSpeed(StrEnum):
    NO_SUCTION = "No_suction"
    QUIET = "Quiet"
    PURE = "Pure"
    STANDARD = "Standard"
    BOOST_IQ = "Boost_IQ"
    TURBO = "Turbo"
    MAX = "Max"
